@@ -1,6 +1,4 @@
-# vim: tabstop=4 shiftwidth=4 softtabstop=4
-
-# Copyright 2015-2016 Lenovo
+# Copyright 2015-2017 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,19 +14,17 @@
 
 import base64
 import binascii
+import socket
+import struct
 import traceback
-import urllib
+import weakref
 
 import pyghmi.constants as pygconst
 import pyghmi.exceptions as pygexc
 import pyghmi.ipmi.oem.generic as generic
-import pyghmi.ipmi.private.constants as ipmiconst
-import pyghmi.ipmi.private.util as util
-
 from pyghmi.ipmi.oem.lenovo import cpu
 from pyghmi.ipmi.oem.lenovo import dimm
 from pyghmi.ipmi.oem.lenovo import drive
-
 from pyghmi.ipmi.oem.lenovo import firmware
 from pyghmi.ipmi.oem.lenovo import imm
 from pyghmi.ipmi.oem.lenovo import inventory
@@ -37,13 +33,17 @@ from pyghmi.ipmi.oem.lenovo import pci
 from pyghmi.ipmi.oem.lenovo import psu
 from pyghmi.ipmi.oem.lenovo import raid_controller
 from pyghmi.ipmi.oem.lenovo import raid_drive
-
-
+import pyghmi.ipmi.private.constants as ipmiconst
+import pyghmi.ipmi.private.util as util
+import pyghmi.redfish.command as redfishcmd
+from pyghmi.redfish.oem.lenovo import tsma
 import pyghmi.util.webclient as wc
 
-import socket
-import struct
-import weakref
+try:
+    from urllib import urlencode
+except ImportError:
+    from urllib.parse import urlencode
+
 try:
     range = xrange
 except NameError:
@@ -143,11 +143,28 @@ class OEMHandler(generic.OEMHandler):
         # will need to retain data to differentiate
         # variations.  For example System X versus Thinkserver
         self.oemid = oemid
+        self._fpc_variant = None
         self.ipmicmd = weakref.proxy(ipmicmd)
         self._has_megarac = None
         self.oem_inventory_info = None
         self._mrethidx = None
         self._hasimm = None
+        self._hasxcc = None
+        if self.has_xcc:
+            self.immhandler = imm.XCCClient(ipmicmd)
+        elif self.has_imm:
+            self.immhandler = imm.IMMClient(ipmicmd)
+        elif self.is_fpc:
+            self.smmhandler = nextscale.SMMClient(ipmicmd)
+        elif self.has_tsma:
+            conn = wc.SecureHTTPConnection(
+                ipmicmd.bmc, 443,
+                verifycallback=self.ipmicmd.certverify)
+            self.tsmahandler = tsma.TsmHandler(None, None, conn,
+                                               fish=redfishcmd)
+            self.tsmahandler.set_credentials(
+                ipmicmd.ipmi_session.userid.decode('utf-8'),
+                ipmicmd.ipmi_session.password.decode('utf-8'))
 
     @property
     def _megarac_eth_index(self):
@@ -157,6 +174,31 @@ class OEMHandler(generic.OEMHandler):
             self._mrethidx = rsp['data'][0]
         return self._mrethidx
 
+    def remove_storage_configuration(self, cfgspec):
+        if self.has_xcc:
+            return self.immhandler.remove_storage_configuration(cfgspec)
+        return super(OEMHandler, self).remove_storage_configuration()
+
+    def clear_storage_arrays(self):
+        if self.has_xcc:
+            return self.immhandler.clear_storage_arrays()
+        return super(OEMHandler, self).clear_storage_ararys()
+
+    def apply_storage_configuration(self, cfgspec):
+        if self.has_xcc:
+            return self.immhandler.apply_storage_configuration(cfgspec)
+        return super(OEMHandler, self).apply_storage_configuration()
+
+    def check_storage_configuration(self, cfgspec):
+        if self.has_xcc:
+            return self.immhandler.check_storage_configuration(cfgspec)
+        return super(OEMHandler, self).get_storage_configuration()
+
+    def get_storage_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.get_storage_configuration()
+        return super(OEMHandler, self).get_storage_configuration()
+
     def get_video_launchdata(self):
         if self.has_tsm:
             return self.get_tsm_launchdata()
@@ -207,12 +249,12 @@ class OEMHandler(generic.OEMHandler):
                     event['severity'] = pygconst.Health.Ok
                     event['component'] = 'User Privilege'
                     event['component_type'] = ipmiconst.sensor_type_codes[6]
-                    event['event_data'] = \
-                        'User {0} on channel {1} had privilege changed ' \
+                    event['event_data'] = (
+                        'User {0} on channel {1} had privilege changed '
                         'from {2} to {3}'.format(
                             oemdata[2], oemdata[1], oemdata[3] & 0b1111,
-                            (oemdata[3] & 0b11110000) >> 4
-                        )
+                            (oemdata[3] & 0b11110000) >> 4)
+                    )
             else:
                 event['event'] = 'OEM event: {0}'.format(
                     ' '.join(format(x, '02x') for x in event['oemdata']))
@@ -234,15 +276,26 @@ class OEMHandler(generic.OEMHandler):
                 event['severity'] = pygconst.Health.Critical
         # For HDD bay events, the event data 2 is the bay, modify
         # the description to be more specific
-        if (event['event_type_byte'] == 0x6f and
-                (evdata[0] & 0b11000000) == 0b10000000 and
-                event['component_type_id'] == 13):
+        if (event['event_type_byte'] == 0x6f
+                and (evdata[0] & 0b11000000) == 0b10000000
+                and event['component_type_id'] == 13):
             event['component'] += ' {0}'.format(evdata[1] & 0b11111)
 
+    def reseat_bay(self, bay):
+        if self.is_fpc:
+            return self.smmhandler.reseat_bay(bay)
+        elif self.has_xcc and bay == -1:
+            return self.immhandler.reseat()
+        return super(OEMHandler, self).reseat_bay(bay)
+
     def get_ntp_enabled(self):
         if self.has_tsm:
             ntpres = self.ipmicmd.xraw_command(netfn=0x32, command=0xa7)
             return ntpres['data'][0] == '\x01'
+        elif self.is_fpc:
+            return self.smmhandler.get_ntp_enabled(self._fpc_variant)
+        elif self.has_tsma:
+            return self.tsmahandler.get_ntp_enabled()
         return None
 
     def get_ntp_servers(self):
@@ -252,6 +305,10 @@ class OEMHandler(generic.OEMHandler):
             srvs.append(ntpres['data'][1:129].rstrip('\x00'))
             srvs.append(ntpres['data'][129:257].rstrip('\x00'))
             return srvs
+        if self.is_fpc:
+            return self.smmhandler.get_ntp_servers()
+        if self.has_tsma:
+            return self.tsmahandler.get_ntp_servers()
         return None
 
     def set_ntp_enabled(self, enabled):
@@ -263,6 +320,11 @@ class OEMHandler(generic.OEMHandler):
                 self.ipmicmd.xraw_command(
                     netfn=0x32, command=0xa8, data=(3, 0), timeout=15)
             return True
+        if self.is_fpc:
+            self.smmhandler.set_ntp_enabled(enabled)
+            return True
+        if self.has_tsma:
+            self.tsmahandler.set_ntp_enabled(enabled)
         return None
 
     def set_ntp_server(self, server, index=0):
@@ -273,15 +335,46 @@ class OEMHandler(generic.OEMHandler):
             cmddata += server.ljust(128, '\x00')
             self.ipmicmd.xraw_command(netfn=0x32, command=0xa8, data=cmddata)
             return True
+        elif self.is_fpc:
+            if not 0 <= index <= 2:
+                raise pygexc.InvalidParameterValue(
+                    'SMM supports indexes 0 through 2')
+            self.smmhandler.set_ntp_server(server, index)
+            return True
+        elif self.has_tsma:
+            if not (0 <= index <= 1):
+                raise pygexc.InvalidParameterValue("Index must be 0 or 1")
+            return self.tsmahandler.set_ntp_server(server, index)
         return None
 
+    def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg,
+                        privilege_level):
+        if self.is_fpc and self._fpc_variant == 2:
+            self.smmhandler.set_user_priv(uid, privilege_level)
+
     @property
     def is_fpc(self):
-        """True if the target is a Lenovo nextscale fan power controller
-        """
-        fpc_ids = ((19046, 32, 1063),)
-        return (self.oemid['manufacturer_id'], self.oemid['device_id'],
-                self.oemid['product_id']) in fpc_ids
+        """True if the target is a Lenovo nextscale fan power controller"""
+
+        if self.has_imm or self.has_xcc:
+            return None
+        if self._fpc_variant is not None:
+            return self._fpc_variant
+        fpc_ids = ((19046, 32, 1063), (20301, 32, 462))
+        smm_id = (19046, 32, 1180)
+        smmv2_2u = (19046, 32, 1360)
+        smmv2_6u = (19046, 32, 1387)
+        currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
+                  self.oemid['product_id'])
+        if currid in fpc_ids:
+            self._fpc_variant = 6
+        elif currid == smm_id:
+            self._fpc_variant = 2
+        elif currid == smmv2_2u:
+            self._fpc_variant = 0x22
+        elif currid == smmv2_6u:
+            self._fpc_variant = 0x26
+        return self._fpc_variant
 
     @property
     def is_sd350(self):
@@ -289,12 +382,20 @@ class OEMHandler(generic.OEMHandler):
                                       self.oemid['device_id'],
                                       self.oemid['product_id'])
 
+    tsma_ids = ((19046, 32, 1287),)
+
+    @property
+    def has_tsma(self):
+        currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
+                  self.oemid['product_id'])
+        return currid in self.tsma_ids
+
     @property
     def has_tsm(self):
-        """True if this particular server have a TSM based service processor
-        """
-        if (self.oemid['manufacturer_id'] == 19046 and
-                self.oemid['device_id'] == 32):
+        """True if this particular server have a TSM based service processor"""
+
+        if (self.oemid['manufacturer_id'] == 19046
+                and self.oemid['device_id'] == 32):
             try:
                 self.ipmicmd.xraw_command(netfn=0x3a, command=0xf)
             except pygexc.IpmiException as ie:
@@ -311,7 +412,7 @@ class OEMHandler(generic.OEMHandler):
                 self._collect_tsm_inventory()
             return iter(self.oem_inventory_info)
         elif self.has_imm:
-            return imm.get_hw_descriptions(self.ipmicmd, self._certverify)
+            return self.immhandler.get_hw_descriptions()
         return ()
 
     def get_oem_inventory(self):
@@ -320,22 +421,33 @@ class OEMHandler(generic.OEMHandler):
             for compname in self.oem_inventory_info:
                 yield (compname, self.oem_inventory_info[compname])
         elif self.has_imm:
-            for inv in imm.get_hw_inventory(self.ipmicmd, self._certverify):
+            for inv in self.immhandler.get_hw_inventory():
                 yield inv
 
     def get_sensor_data(self):
-        if self.is_fpc:
-            for name in nextscale.get_sensor_names():
-                yield nextscale.get_sensor_reading(name, self.ipmicmd)
+        if self.has_imm:
+            for name in self.immhandler.get_oem_sensor_names(self.ipmicmd):
+                yield self.immhandler.get_oem_sensor_reading(name,
+                                                             self.ipmicmd)
+        elif self.is_fpc:
+            for name in nextscale.get_sensor_names(self._fpc_variant):
+                yield nextscale.get_sensor_reading(name, self.ipmicmd,
+                                                   self._fpc_variant)
 
     def get_sensor_descriptions(self):
-        if self.is_fpc:
-            return nextscale.get_sensor_descriptions()
+        if self.has_imm:
+            return self.immhandler.get_oem_sensor_descriptions(self.ipmicmd)
+        elif self.is_fpc:
+            return nextscale.get_sensor_descriptions(self._fpc_variant)
         return ()
 
     def get_sensor_reading(self, sensorname):
-        if self.is_fpc:
-            return nextscale.get_sensor_reading(sensorname, self.ipmicmd)
+        if self.has_imm:
+            return self.immhandler.get_oem_sensor_reading(sensorname,
+                                                          self.ipmicmd)
+        elif self.is_fpc:
+            return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
+                                                self._fpc_variant)
         return ()
 
     def get_inventory_of_component(self, component):
@@ -343,8 +455,7 @@ class OEMHandler(generic.OEMHandler):
             self._collect_tsm_inventory()
             return self.oem_inventory_info.get(component, None)
         if self.has_imm:
-            return imm.get_component_inventory(self.ipmicmd, self._certverify,
-                                               component)
+            return self.immhandler.get_component_inventory(component)
 
     def _collect_tsm_inventory(self):
         self.oem_inventory_info = {}
@@ -416,7 +527,7 @@ class OEMHandler(generic.OEMHandler):
         else:
             raise pygexc.UnsupportedFunctionality()
 
-    def process_fru(self, fru):
+    def process_fru(self, fru, name=None):
         if fru is None:
             return fru
         if self.has_tsm:
@@ -462,7 +573,11 @@ class OEMHandler(generic.OEMHandler):
                 endidx = len(macs) - 5
                 macprefix = None
                 while idx < endidx:
-                    currmac = macs[idx:idx+6]
+                    currmac = macs[idx:idx + 6]
+                    if not isinstance(currmac, bytearray):
+                        # invalid vpd format, abort attempts to extract
+                        # mac in this way
+                        break
                     if currmac == b'\x00\x00\x00\x00\x00\x00':
                         break
                     # VPD may veer off, detect and break off
@@ -476,11 +591,40 @@ class OEMHandler(generic.OEMHandler):
                     idx = idx + 6
             except (AttributeError, KeyError, IndexError):
                 pass
+            if self.has_xcc and name and name.startswith('PSU '):
+                self.immhandler.augment_psu_info(fru, name)
             return fru
+        elif self.is_fpc and self.is_fpc != 6:  # SMM variant
+            fru['oem_parser'] = 'lenovo'
+            return self.smmhandler.process_fru(fru)
         else:
             fru['oem_parser'] = None
             return fru
 
+    @property
+    def has_xcc(self):
+        if self._hasxcc is not None:
+            return self._hasxcc
+        try:
+            bdata = self.ipmicmd.xraw_command(netfn=0x3a, command=0xc1)
+        except pygexc.IpmiException:
+            self._hasxcc = False
+            self._hasimm = False
+            return False
+        if len(bdata['data'][:]) != 3:
+            self._hasimm = False
+            self._hasxcc = False
+            return False
+        rdata = bytearray(bdata['data'][:])
+        self._hasxcc = rdata[1] & 16 == 16
+        if self._hasxcc:
+            # For now, have imm calls go to xcc, since they are providing same
+            # interface.  Longer term the hope is that all the Lenovo
+            # stuff will branch at init, and not have conditionals
+            # in all the functions
+            self._hasimm = self._hasxcc
+        return self._hasxcc
+
     @property
     def has_imm(self):
         if self._hasimm is not None:
@@ -497,15 +641,31 @@ class OEMHandler(generic.OEMHandler):
         self._hasimm = (rdata[1] & 1 == 1) or (rdata[1] & 16 == 16)
         return self._hasimm
 
-    def get_oem_firmware(self, bmcver):
+    def get_oem_firmware(self, bmcver, components):
         if self.has_tsm:
             command = firmware.get_categories()["firmware"]
             rsp = self.ipmicmd.xraw_command(**command["command"])
             return command["parser"](rsp["data"])
         elif self.has_imm:
-            return imm.get_firmware_inventory(self.ipmicmd, bmcver,
-                                              self._certverify)
-        return super(OEMHandler, self).get_oem_firmware(bmcver)
+            return self.immhandler.get_firmware_inventory(bmcver, components)
+        elif self.is_fpc:
+            return nextscale.get_fpc_firmware(bmcver, self.ipmicmd,
+                                              self._fpc_variant)
+        elif self.has_tsma:
+            return self.tsmahandler.get_firmware_inventory(
+                components, raisebypass=False, ipmicmd=self.ipmicmd)
+        return super(OEMHandler, self).get_oem_firmware(bmcver, components)
+
+    def get_diagnostic_data(self, savefile, progress, autosuffix=False):
+        if self.has_xcc:
+            return self.immhandler.get_diagnostic_data(savefile, progress,
+                                                       autosuffix)
+        if self.is_fpc:
+            return self.smmhandler.get_diagnostic_data(savefile, progress,
+                                                       autosuffix)
+        if self.has_tsma:
+            return self.tsmahandler.get_diagnostic_data(savefile, progress,
+                                                        autosuffix)
 
     def get_oem_capping_enabled(self):
         if self.has_tsm:
@@ -552,6 +712,8 @@ class OEMHandler(generic.OEMHandler):
                                                 data=(4, i))
                 name += rsp['data'][:]
             return name.rstrip('\x00')
+        elif self.is_fpc:
+            return self.smmhandler.get_domain()
 
     def set_oem_domain_name(self, name):
         if self.has_tsm:
@@ -562,13 +724,29 @@ class OEMHandler(generic.OEMHandler):
             # set the domain name content
             name = name.ljust(256, "\x00")
             for i in range(0, 4):
-                data = [4, i+1]
-                offset = i*64
-                data.extend([ord(x) for x in name[offset:offset+64]])
+                data = [4, i + 1]
+                offset = i * 64
+                data.extend([ord(x) for x in name[offset:offset + 64]])
                 self.ipmicmd.xraw_command(netfn=0x32, command=0x6c, data=data)
 
             self._restart_dns()
             return
+        elif self.is_fpc:
+            self.smmhandler.set_domain(name)
+
+    def set_hostname(self, hostname):
+        if self.has_xcc:
+            return self.immhandler.set_hostname(hostname)
+        elif self.is_fpc:
+            return self.smmhandler.set_hostname(hostname)
+        return super(OEMHandler, self).set_hostname(hostname)
+
+    def get_hostname(self):
+        if self.has_xcc:
+            return self.immhandler.get_hostname()
+        elif self.is_fpc:
+            return self.smmhandler.get_hostname()
+        return super(OEMHandler, self).get_hostname()
 
     """ Gets a remote console launcher for a Lenovo ThinkServer.
 
@@ -577,12 +755,12 @@ class OEMHandler(generic.OEMHandler):
     def _get_ts_remote_console(self, bmc, username, password):
         # We don't establish non-secure connections without checking
         # certificates
-        if not self._certverify:
+        if not self.ipmicmd.certverify:
             return
         conn = wc.SecureHTTPConnection(bmc, 443,
-                                       verifycallback=self._certverify)
+                                       verifycallback=self.ipmicmd.certverify)
         conn.connect()
-        params = urllib.urlencode({
+        params = urlencode({
             'WEBVAR_USERNAME': username,
             'WEBVAR_PASSWORD': password
         })
@@ -639,7 +817,7 @@ class OEMHandler(generic.OEMHandler):
             else:
                 # fall back to a dumber, but more universal formatter
                 ipv6str = binascii.b2a_hex(ipv6_addr)
-                ipv6str = ':'.join([ipv6str[x:x+4] for x in range(0, 32, 4)])
+                ipv6str = ':'.join([ipv6str[x:x + 4] for x in range(0, 32, 4)])
             netdata['ipv6_addresses'] = [
                 '{0}/{1}'.format(ipv6str, ipv6_prefix)]
 
@@ -698,7 +876,7 @@ class OEMHandler(generic.OEMHandler):
                                   data=(1, selector, 0, 1))
         # now do the set
         for x in range(0, 256, 64):
-            currdata = padded[x:x+64]
+            currdata = padded[x:x + 64]
             currchunk = x // 64 + 1
             cmddata = [1, selector, currchunk] + currdata
             self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=cmddata)
@@ -712,7 +890,7 @@ class OEMHandler(generic.OEMHandler):
         imgnames = rsp['data'][1:]
         shortnames = []
         for idx in range(0, len(imgnames), 22):
-            shortnames.append(imgnames[idx+2:idx+22].rstrip('\0'))
+            shortnames.append(imgnames[idx + 2:idx + 22].rstrip('\0'))
         return shortnames
 
     def _megarac_media_waitforready(self, imagename):
@@ -763,8 +941,10 @@ class OEMHandler(generic.OEMHandler):
 
     def attach_remote_media(self, url, username, password):
         if self.has_imm:
-            imm.attach_remote_media(self.ipmicmd, self._certverify, url,
-                                    username, password)
+            self.immhandler.attach_remote_media(url, username, password)
+        elif self.has_tsma:
+            return self.tsmahandler.attach_remote_media(
+                url, username, password, None)
         elif self.has_megarac:
             proto, host, path = util.urlsplit(url)
             if proto == 'smb':
@@ -786,10 +966,128 @@ class OEMHandler(generic.OEMHandler):
                 else:
                     raise
 
+    def update_firmware(self, filename, data=None, progress=None, bank=None):
+        if self.has_xcc:
+            return self.immhandler.update_firmware(
+                filename, data=data, progress=progress, bank=bank)
+        if self.is_fpc:
+            return self.smmhandler.update_firmware(
+                filename, data=data, progress=progress, bank=bank)
+        if self.has_tsma:
+            return self.tsmahandler.update_firmware(
+                filename, data=data, progress=progress, bank=bank)
+        return super(OEMHandler, self).update_firmware(filename, data=data,
+                                                       progress=progress,
+                                                       bank=bank)
+
+    def get_description(self):
+        if self.has_xcc:
+            return self.immhandler.get_description()
+        if self.is_fpc:
+            return {'height': self._fpc_variant, 'slot': 0}
+        return super(OEMHandler, self).get_description()
+
+    def get_extended_bmc_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.get_extended_bmc_configuration()
+        return super(OEMHandler, self).get_extended_bmc_configuration()
+
+    def get_bmc_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.get_bmc_configuration()
+        if self.is_fpc:
+            return self.smmhandler.get_bmc_configuration()
+        if self.has_tsma:
+            return self.tsmahandler.get_bmc_configuration()
+        return super(OEMHandler, self).get_bmc_configuration()
+
+    def set_bmc_configuration(self, changeset):
+        if self.has_xcc:
+            return self.immhandler.set_bmc_configuration(changeset)
+        if self.is_fpc:
+            return self.smmhandler.set_bmc_configuration(
+                changeset, self._fpc_variant)
+        if self.has_tsma:
+            return self.tsmahandler.set_bmc_configuration(
+                changeset)
+        return super(OEMHandler, self).set_bmc_configuration(changeset)
+
+    def get_system_configuration(self, hideadvanced):
+        if self.has_imm or self.has_xcc:
+            return self.immhandler.get_system_configuration(hideadvanced)
+        if self.has_tsma:
+            return self.tsmahandler.get_uefi_configuration(hideadvanced)
+        return super(OEMHandler, self).get_system_configuration(hideadvanced)
+
+    def set_system_configuration(self, changeset):
+        if self.has_imm or self.has_xcc:
+            return self.immhandler.set_system_configuration(changeset)
+        if self.has_tsma:
+            return self.tsmahandler.set_uefi_configuration(changeset)
+        return super(OEMHandler, self).set_system_configuration(changeset)
+
+    def clear_bmc_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.clear_bmc_configuration()
+        elif self.is_fpc:
+            return self.smmhandler.clear_bmc_configuration()
+        return super(OEMHandler, self).clear_system_configuration()
+
+    def clear_system_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.clear_system_configuration()
+        if self.has_tsma:
+            return self.tsmahandler.clear_uefi_configuration()
+        return super(OEMHandler, self).clear_system_configuration()
+
     def detach_remote_media(self):
         if self.has_imm:
-            imm.detach_remote_media(self.ipmicmd, self._certverify)
+            self.immhandler.detach_remote_media()
+        elif self.has_tsma:
+            self.tsmahandler.detach_remote_media()
         elif self.has_megarac:
             self.ipmicmd.xraw_command(
                 netfn=0x32, command=0x9f, data=(8, 10, 0, 0))
             self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 11))
+
+    def upload_media(self, filename, progress, data):
+        if self.has_xcc or self.has_imm:
+            return self.immhandler.upload_media(filename, progress, data)
+        return super(OEMHandler, self).upload_media(filename, progress, data)
+
+    def list_media(self):
+        if self.has_xcc or self.has_imm:
+            return self.immhandler.list_media()
+        if self.has_tsma:
+            return self.tsmahandler.list_media()
+        return super(OEMHandler, self).list_media()
+
+    def get_health(self, summary):
+        if self.has_xcc:
+            return self.immhandler.get_health(summary)
+        return super(OEMHandler, self).get_health(summary)
+
+    def get_licenses(self):
+        if self.has_xcc:
+            return self.immhandler.get_licenses()
+        return super(OEMHandler, self).get_licenses()
+
+    def get_user_expiration(self, uid):
+        if self.has_xcc:
+            return self.immhandler.get_user_expiration(uid)
+        return None
+
+    def delete_license(self, name):
+        if self.has_xcc:
+            return self.immhandler.delete_license(name)
+        return super(OEMHandler, self).delete_license(name)
+
+    def save_licenses(self, directory):
+        if self.has_xcc:
+            return self.immhandler.save_licenses(directory)
+        return super(OEMHandler, self).save_licenses(directory)
+
+    def apply_license(self, filename, progress=None, data=None):
+        if self.has_xcc:
+            return self.immhandler.apply_license(filename, progress, data)
+        return super(OEMHandler, self).apply_license(filename, progress, data)
